/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  libgrg_crypt_v1.c - routines for data encryption and writing according to file
 *                   	format v.1, provided for compatibility (read only). This
 *						format was used only by Gringotts, so the header and some
 *						other parameters refer to that program.
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <mcrypt.h>
#include <mhash.h>
#include <zlib.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "libgrg_crypt.h"
#include "libgrg_utils.h"
#include "libgringotts.h"

//only version 1 used only these
#define GRG_KEY_SIZE			32	//256 bit
#define	GRG_FLAG				"GRG"
#define GRG_FLAG_LEN			3
#define GRG_OVERHEAD_V1			GRG_FLAG_LEN+GRG_FILE_VERSION_LEN+GRG_CRC_LEN*2+GRG_ALGO_LEN+GRG_DATA_DIM_LEN

/**
 * get_key_v1:
 * @pwd: the password to generate hash key from
 *
 * Produces a key, for use with libmcrypt's encryption engine, using the
 * 256-bit SHA hash of a password
 *
 * Returns: a string (GRG_KEY_SIZE byte) with the key
 */
static unsigned char *
get_key_v1 (unsigned char *pwd, unsigned long pwdlen)
{
	unsigned char *ret;

	ret = calloc (1, GRG_KEY_SIZE);

	mhash_keygen (KEYGEN_MCRYPT, MHASH_SHA256, 0, ret, GRG_KEY_SIZE, NULL,
		      0, pwd, pwdlen);

	return ret;
}

/**
 * get_IV_v1:
 * @IV_size: the size of the IV to produce
 * @salt: the salt to produce the IV with
 * @salt_size: the size of the salt
 * @pwd: the password of the encrypted data
 *
 * Generates an `IV' token. See libmcrypt docs to know what it is... ;)
 *
 * Returns: a byte sequence with the IV
 */
static unsigned char *
get_IV_v1 (unsigned int IV_size, unsigned char *salt,
	unsigned int salt_size, unsigned char *pwd, unsigned long pwdlen)
{
	unsigned char *ret;

	ret = calloc (1, IV_size);

	mhash_keygen (KEYGEN_S2K_SALTED, MHASH_GOST, 0, ret, IV_size, salt,
		      salt_size, pwd, pwdlen);

	return ret;
}

extern int
compare_CRC32 (unsigned char *, unsigned char *, unsigned long);

/**
 * grg_load_crypted_v1:
 * @origData: a pointer to a byte sequence to store the data in. It must be freed after use!
 * @pwd: the password to decode data
 * @fpath: path of the file to read data from
 * @just_validate: if 1, it only checks for file validity, without loading it.
 * 
 * Loads the encrypted data stored in a Gringotts file (v.1), or just check for validity
 *
 * Returns: 0 if OK; an error code otherwise (see grg_crypt.h)
 */
int
grg_load_crypted_v1 (GRG_TOKEN gt, unsigned char **origData, unsigned long *odlen,
		  unsigned char *pwd, unsigned int pwdlen,
		  const unsigned char *fpath, unsigned int just_validate)
{
	unsigned char *CRC32, *CRC32b, *compData, *bcompData, *IV,
		*key, *dimdata, *salt, *tmpData;
	unsigned int err, dIV, salt_size;
	long flen, bflen, i;
	FILE *fp;
	MCRYPT mod;

	salt_size = mhash_get_keygen_salt_size (KEYGEN_S2K_SALTED);

	fp = fopen (fpath, "rb");

	if (!fp)
		return GRG_READ_FILE_ERR;

	flen = grg_file_size (fp);

	if (flen < GRG_OVERHEAD_V1 + salt_size)	//the total length of header part
		return GRG_READ_FILE_ERR;

	if (flen - GRG_OVERHEAD_V1 - salt_size > MAX_COMP_SIZE * 1048576l)
		return GRG_READ_TOO_BIG_ERR;

	//skip the already checked part

	fseek (fp, GRG_FLAG_LEN + GRG_FILE_VERSION_LEN, SEEK_SET);

	//checks the 1st CRC

	flen -= GRG_FLAG_LEN + GRG_FILE_VERSION_LEN + GRG_CRC_LEN;

	CRC32 = (unsigned char *) malloc (GRG_CRC_LEN *
					  sizeof (unsigned char));
	compData = (unsigned char *) malloc (flen * sizeof (unsigned char));
	bcompData = compData;
	
	bflen = flen;

	fread (CRC32, 1, GRG_CRC_LEN, fp);
	fread (compData, 1, flen, fp);

	fclose (fp);

	if (!compare_CRC32 (CRC32, compData, flen))
	{
		grg_free (gt, CRC32, GRG_CRC_LEN);
		CRC32 = NULL;
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		return GRG_READ_CRC_ERR;
	}

	grg_free (gt, CRC32, GRG_CRC_LEN);
	CRC32 = NULL;

	if (just_validate == 1)
		return GRG_OK;

	//don't care ALGO, for now
	compData += GRG_ALGO_LEN * sizeof (unsigned char);
	flen -= GRG_ALGO_LEN * sizeof (unsigned char);

	//decrypts the encrypted data
	mod = mcrypt_module_open (MCRYPT_SERPENT, NULL, MCRYPT_CFB, NULL);

	if (mod == MCRYPT_FAILED)
	{
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		return GRG_READ_ENC_INIT_ERR;
	}

	key = get_key_v1 (pwd, pwdlen);
	dIV = mcrypt_enc_get_iv_size (mod);
	salt = grg_strdup (compData, salt_size);
	IV = get_IV_v1 (dIV, salt, salt_size, pwd, pwdlen);

	mcrypt_generic_init (mod, key, GRG_KEY_SIZE, IV);
	grg_free (gt, key, GRG_KEY_SIZE);
	key = NULL;
	grg_free (gt, IV, dIV);
	IV = NULL;
	grg_free (gt, salt, salt_size);
	salt = NULL;
	if (err < 0)
	{
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		return GRG_READ_ENC_INIT_ERR;
	}

	compData += salt_size * sizeof (unsigned char);
	flen -= salt_size;

	for (i = 0; i < flen; i++)
		mdecrypt_generic (mod, &compData[i], 1);

	mcrypt_generic_deinit (mod);
	mcrypt_module_close (mod);

	//checks the 2nd CRC32

	CRC32b = grg_strdup (compData, GRG_CRC_LEN);

	compData += GRG_CRC_LEN * sizeof (unsigned char);
	flen -= GRG_CRC_LEN * sizeof (unsigned char);

	if (!compare_CRC32 (CRC32b, compData, flen))
	{
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		grg_free (gt, CRC32b, GRG_CRC_LEN);
		CRC32b = NULL;
		return GRG_READ_PWD_ERR;
	}

	grg_free (gt, CRC32b, GRG_CRC_LEN);
	CRC32b = NULL;

	//reads the uncompressed data length

	dimdata = grg_strdup (compData, GRG_DATA_DIM_LEN);

	compData += GRG_DATA_DIM_LEN * sizeof (unsigned char);
	flen -= GRG_DATA_DIM_LEN * sizeof (unsigned char);

	*odlen = grg_char2long (dimdata);

	grg_free (gt, dimdata, GRG_DATA_DIM_LEN);
	dimdata = NULL;

	//uncompress the final data

	tmpData =
		(unsigned char *) malloc ((*odlen) * sizeof (unsigned char));

	err = uncompress (tmpData, odlen, compData, flen);
	
	
	if(err<0)
	{
		grg_free (gt, bcompData, bflen);
		bcompData = NULL;
		grg_free (gt, tmpData, *odlen);
		tmpData = NULL;
		return GRG_READ_COMP_ERR;
	}

	*origData = grg_strconcat(tmpData, "", NULL, *odlen, 1, 0);
	
	grg_free (gt, bcompData, bflen);
	bcompData = NULL;
	grg_free (gt, tmpData, *odlen);
	tmpData = NULL;
	
	return GRG_OK;
}
