/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_entries.c - functions to manage the GList of Gringotts' entries
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <stdlib.h>
#include <string.h>

#include "../libgringotts/libgringotts.h"

#include "grg_defs.h"
#include "grg_utils.h"
#include "gringotts.h"
#include "grg_prefs.h"
#include "grg_safe.h"
#include "grg_recent_dox.h"

struct grg_entry
{
	gchar *entryID;
	gchar *entryBody;
};

static GList *entries = NULL;
static GList *current = NULL;
static guchar *serialized;

/**
 * meta_free:
 * @data: the callback's data
 * @user_data: the callback's user-defined data
 *
 * Frees a single node. Used only by grg_entries_free()
 */
static void
meta_free (gpointer data, gpointer user_data)
{
	struct grg_entry *entry = (struct grg_entry *) data;

	grg_free (gt, entry->entryID, -1);
	entry->entryID = NULL;
	grg_free (gt, entry->entryBody, -1);
	entry->entryBody = NULL;
	g_free (entry);
	entry = NULL;
}

/**
 * grg_entries_append:
 * @ID: the title of the new entry
 *
 * Appends a new entry to the list
 */
void
grg_entries_append (gchar * ID)
{
	struct grg_entry *entry;

	entry = g_malloc (sizeof (struct grg_entry));

	entry->entryID = g_strdup (ID);
	entry->entryBody = g_strdup ("");

	entries = g_list_append (entries, entry);
	current = g_list_last (entries);
}

/**
 * grg_entries_remove:
 *
 * Removes the current entry and frees it
 */
void
grg_entries_remove (void)
{
	gint pos;

	if (!current)
		return;

	pos = g_list_position (entries, current);
	meta_free (current->data, NULL);
	entries = g_list_remove_link (entries, current);
	g_list_free_1 (current);
	current = g_list_nth (entries, pos);
	if (!current)
		current = g_list_nth (entries, pos - 1);
}

/**
 * grg_entries_is_first:
 *
 * Tells if the current entry is the first one
 *
 * Returns: TRUE if true, FALSE if false.
 */
gboolean
grg_entries_is_first (void)
{
	if (!current)
		return TRUE;
	if (!g_list_previous (current))
		return TRUE;
	return FALSE;
}

/**
 * grg_entries_is_last:
 *
 * Tells if the current entry is the last one
 *
 * Returns: TRUE if true, FALSE if false.
 */
gboolean
grg_entries_is_last (void)
{
	if (!current)
		return TRUE;

	if (!g_list_next (current))
		return TRUE;

	return FALSE;
}

/**
 * grg_entries_is_empty:
 *
 * Tells if the current list is empty
 *
 * Returns: TRUE if true, FALSE if false.
 */
gboolean
grg_entries_is_empty (void)
{
	return (entries == NULL);
}

/**
 * grg_entries_first:
 *
 * Goes to the first entry
 */
void
grg_entries_first (void)
{
	current = g_list_first (entries);
}

/**
 * grg_entries_prev:
 *
 * Goes to the previous entry
 */
void
grg_entries_prev (void)
{
	if (!grg_entries_is_first ())
		current = g_list_previous (current);
}

/**
 * grg_entries_next:
 *
 * Goes to the next entry
 */
void
grg_entries_next (void)
{
	if (!grg_entries_is_last ())
		current = g_list_next (current);
}

/**
 * grg_entries_last:
 *
 * Goes to the last entry
 */
void
grg_entries_last (void)
{
	current = g_list_last (entries);
}

/**
 * grg_entries_nth:
 * @pos: the position to go to
 *
 * Goes to the specified entry, if possible
 */
void
grg_entries_nth (gint pos)
{
	current = g_list_nth (entries, pos);
}

/**
 * grg_entries_position:
 *
 * Tells the position of the current entry in the list
 *
 * Returns: a gint with the position
 */
gint
grg_entries_position (void)
{
	if (!current)
		return -1;

	return g_list_position (entries, current);
}

/**
 * grg_entries_raise
 *
 * Shifts one position up
 */
void
grg_entries_raise (void)
{
	gpointer ent;

	if (!current->prev)
		return;

	ent = current->data;
	current->data = current->prev->data;
	current->prev->data = ent;
}

/**
 * grg_entries_sink
 *
 * Shifts one position down
 */
void
grg_entries_sink (void)
{
	gpointer ent;

	if (!current->next)
		return;

	ent = current->data;
	current->data = current->next->data;
	current->next->data = ent;
}

/**
 * grg_entries_get_ID
 *
 * Returns the title of the current entry.
 *
 * Returns: a gchar* with the title (NOT a copy), or NULL if the list is empty
 */
gchar *
grg_entries_get_ID (void)
{
	if (entries)
		return ((struct grg_entry *) current->data)->entryID;
	else
		return NULL;
}

/**
 * grg_entries_get_Body
 *
 * Returns the body text of the current entry.
 *
 * Returns: a gchar* with the text, or NULL if the list is empty
 */
gchar *
grg_entries_get_Body (void)
{
	if (entries)
		return ((struct grg_entry *) current->data)->entryBody;
	else
		return NULL;
}

/**
 * grg_entries_set_ID
 * @ID: the text to store
 *
 * Stores the given title in the current entry.
 */
void
grg_entries_set_ID (gchar * ID)
{
	if(current)
		((struct grg_entry *) current->data)->entryID = g_strdup (ID);
}

/**
 * grg_entries_set_Body
 * @Body: the text to store
 *
 * Stores the given text as the body of the current entry.
 */
void
grg_entries_set_Body (gchar * Body)
{
	if(current)
		((struct grg_entry *) current->data)->entryBody = g_strdup (Body);
}

/**
 * grg_entries_free:
 *
 * Deletes and frees all the list
 */
void
grg_entries_free (void)
{
	g_list_foreach (entries, meta_free, NULL);
	g_list_free (entries);
	entries = NULL;
	current = NULL;
}

/**
 * meta_save:
 * @data: the callback's data
 * @user_data: the callback's user-defined data
 *
 * "serializes" a single node. Used only by grg_entries_save()
 */
static void
meta_save (gpointer data, gpointer user_data)
{
	struct grg_entry *entry = (struct grg_entry *) data;
	gchar *eBody, *res;

	eBody = g_strescape (entry->entryBody, "");

	res = g_strconcat (serialized, "\n", entry->entryID, "\n", eBody,
			   NULL);

	grg_free (gt, eBody, -1);
	eBody = NULL;

	grg_free (gt, serialized, -1);
	serialized = g_strdup (res);

	grg_free (gt, res, -1);
	res = NULL;
}

/**
 * grg_entries_save:
 * @file: the path of the file to save
 * @pwd: the password to use
 *
 * Saves the list into an encrypted file
 *
 * Returns: GRG_OK if all is well, an error if not
 */
gint
grg_entries_save (gchar * file, gchar * pwd, gint pwd_len)
{
	gint err;

	serialized = g_strdup ("");
	g_list_foreach (entries, meta_save, NULL);
	//the "+1" strips off the initial separator

#ifdef PWD_HIDING
	grg_security_unseal (pwd, pwd_len);
#endif
	err = grg_save_crypted (gt, serialized + 1, strlen (serialized) - 1, pwd,
				pwd_len, file);
#ifdef PWD_HIDING
	grg_security_seal (pwd, pwd_len);
#endif

	if (err != GRG_OK)
		return err;

	grg_free (gt, serialized, -1);
	serialized = NULL;

	grg_recent_dox_push (file);
	return GRG_OK;
}

/**
 * grg_entries_load_from_string:
 * @str: the string which contains the data
 *
 * "de-serializes" a string into an entry list
 */
void
grg_entries_load_from_string (gchar * str)
{
	gchar **fields;
	gint num;

	fields = g_strsplit (str, "\n", 0);

	num = 0;
	while (fields[num * 2] != NULL)
	{
		gchar *Body;

		grg_entries_append (fields[num * 2]);
		Body = g_strcompress (fields[num * 2 + 1]);
		grg_entries_set_Body (Body);

		grg_free (gt, Body, -1);
		Body = NULL;
		num++;
	}

	grg_entries_first ();

	g_strfreev (fields);
}

/**
 * grg_load_wrapper:
 * @txt: a pointer to a byte sequence to store the data in. It must be freed after use!
 * @pwd: the password to decode data
 * @file: path of the file to read data from
 * @just_validate: if 1, it only checks for file validity, without loading it.
 * 
 * Wrapper to grg_load_crypted, to add UTF-8 validation.
 *
 * Returns: 0 if OK; an error code otherwise (see grg_crypt.h)
 */
gint
grg_load_wrapper (guchar ** txt, gchar * pwd, gint pwd_len, const gchar * file,
		  guint just_validate)
{
	gint err;
	gulong len = 0;

#ifdef PWD_HIDING
	grg_security_unseal (pwd, pwd_len);
#endif
	err = grg_load_crypted (gt, txt, &len, pwd, pwd_len, file, just_validate);
#ifdef PWD_HIDING
	grg_security_seal (pwd, pwd_len);
#endif

	grg_prefs_update();

	if (err != GRG_OK)
		return err;

	if (just_validate)
		return GRG_OK;

//#ifndef G_DISABLE_ASSERT
//	g_assert (g_utf8_strlen (*txt, -1) == len);
//#endif

	if (!g_utf8_validate (*txt, -1, NULL))
		return GRG_READ_INVALID_CHARSET_ERR;

	grg_recent_dox_push (file);
	return GRG_OK;
}

/**
 * grg_entries_load:
 * @file: the path of the file to load
 * @pwd: the password to use
 *
 * Reads a file into an entry list
 *
 * Returns: GRG_OK if all is well, an error if not
 */
gint
grg_entries_load (gchar * file, gchar * pwd, gint pwd_len)
{
	gint err;

	grg_entries_free ();

	err = grg_load_wrapper (&serialized, pwd, pwd_len, file, 0);
	if (err != GRG_OK)
		return err;

	grg_entries_load_from_string (serialized);

	grg_free (gt, serialized, -1);
	serialized = NULL;

	return GRG_OK;
}

/**
 * grg_entries_n_el:
 *
 * Tells how many elements does the list contain
 *
 * Returns: a guint with the elements' number
 */
guint
grg_entries_n_el (void)
{
	if (entries)
		return g_list_length (entries);
	else
		return 0;
}

/**
 * grg_entries_find:
 * @needle: the text to find
 * @offset: the offset to search from
 * @only_current: if TRUE, consider only the current entry
 * @case_sens: wheter to be case sensitive or not
 *
 * Searches for a text in the entries.
 *
 * Returns: the offset of the found text, -1 if not found
 */
glong
grg_entries_find (gchar * needle, glong offset, gboolean only_current,
		  gboolean case_sens)
{
	guchar *text = grg_entries_get_Body (), *start = g_utf8_offset_to_pointer (text, offset), *occur;
	glong result;

	if (case_sens)
	{
		occur = g_strstr_len (start, strlen (start), needle);

		if (occur)
			result = g_utf8_pointer_to_offset (text,
							   occur) + offset;
		else
			result = -1;
	}
	else
	{
		gchar *tok1, *tok2;
		gulong len;

		tok1 = g_utf8_casefold (start, -1);
		tok2 = g_utf8_casefold (needle, -1);
		len = strlen (tok1);

		occur = g_strstr_len (tok1, len, tok2);

		if (occur)
			result = g_utf8_pointer_to_offset (tok1,
							   occur) + offset;
		else
			result = -1;

		grg_free (gt, tok1, len);
		tok1 = NULL;
		g_free (tok2);
		tok2 = NULL;
	}

	if (result < 0)
	{
		gint cur = grg_entries_position (), ret;

		if (only_current)
			return -1;

		if (grg_entries_is_last ())
			return -1;
		else
			grg_entries_next ();

		ret = grg_entries_find (needle, 0, FALSE, case_sens);
		if (ret == -1)
			grg_entries_nth (cur);
		return ret;
	}

	return result;
}
