/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_prefs_io.c - preferences load/saving
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#include "../libgringotts/libgringotts.h"

#include "grg_prefs.h"
#include "grg_prefs_io.h"
#include "grg_utils.h"
#include "grg_defs.h"
#include "gringotts.h"

gint
grg_save_prefs (void)
{
	guchar *path, *row, algo, *grg_pref_file_local;
	FILE *fp;

	path = g_build_filename (g_get_home_dir (), ".gringotts.conf", NULL);
	fp = fopen (path, "w");
	g_free (path);

	if (fp == NULL)
		return GRG_PREFS_IO_ERROR;

	algo = (guchar) (grg_tok_get_crypt_algo(gt) | grg_tok_get_hash_algo(gt) | grg_tok_get_comp_algo(gt) |
			 grg_tok_get_comp_ratio(gt));

	//writes an header
	row = "<!-- this isn't (yet) true XML. It's *very* positional, so please leave it as it is -->\n";
	fwrite (row, strlen (row), 1, fp);

	//saves the algorythm
	row = g_strdup_printf ("<algo_code>\n%02x\n</algo_code>\n", algo);
	fwrite (row, strlen (row), 1, fp);
	g_free (row);

	//saves the startup file
	grg_pref_file_local = get_pref_file ();
	if(grg_pref_file_local)
	{
		row = g_strdup_printf ("<startup_file>\n%.255s\n</startup_file>\n", grg_pref_file_local);
		g_free (grg_pref_file_local);
	}
	else
		row = g_strdup ("<startup_file>\n\n</startup_file>\n");
	fwrite (row, strlen (row), 1, fp);
	g_free (row);

	fclose (fp);

	return GRG_OK;
}

gint
grg_load_prefs (void)
{
	guchar *path, algo, sfile[256];
	gint algo1, algo2;
	FILE *fp;

	path = g_build_filename (g_get_home_dir (), ".gringotts.conf", NULL);
	fp = fopen (path, "r");
	g_free (path);

	if (fp == NULL)
		return GRG_PREFS_NO_FILE;

	//skip first two lines
	if (!skip_lines (fp, 2))
	{
		fclose (fp);
		return GRG_PREFS_INVALID;
	}

	//reads the two algo chars
	algo1 = fgetc (fp);
	if (algo1 == EOF)
	{
		fclose (fp);
		return GRG_PREFS_INVALID;
	}
	algo2 = fgetc (fp);
	if (algo2 == EOF)
	{
		fclose (fp);
		return GRG_PREFS_INVALID;
	}
	if (!
	    (((algo1 >= '0') && (algo1 <= '9'))
	     || ((algo1 >= 'a') && (algo1 <= 'f'))) || !(((algo2 >= '0')
							  && (algo2 <= '9'))
							 || ((algo2 >= 'a')
							     && (algo2 <=
								 'f'))))
	{
		fclose (fp);
		return GRG_PREFS_INVALID;
	}
	//skip three lines
	if (!skip_lines (fp, 3))
	{
		fclose (fp);
		return GRG_PREFS_INVALID;
	}
	//reads the startup file
	fgets(sfile, 255, fp);
	//skip one other line
	if (!skip_lines (fp, 1))
	{
		fclose (fp);
		return GRG_PREFS_INVALID;
	}
	
	//form the algo byte and setup values
	if (algo1 < 'a')
		algo1 -= '0';
	else
		algo1 -= 'W';	//'a'-10

	if (algo2 < 'a')
		algo2 -= '0';
	else
		algo2 -= 'W';	//'a'-10

	algo = 0;
	algo |= (guchar) ((algo1 << 4) & 0xf0);
	algo |= (guchar) (algo2 & 0x0f);

	grg_tok_set_crypt_algo(gt, (grg_crypt_algo) (algo & GRG_ENCRYPT_MASK));
	grg_tok_set_hash_algo(gt, (grg_hash_algo) (algo & GRG_HASH_MASK));
	grg_tok_set_comp_algo(gt, (grg_comp_algo) (algo & GRG_COMP_TYPE_MASK));
	grg_tok_set_comp_ratio(gt, (grg_comp_ratio) (algo & GRG_COMP_LVL_MASK));
	if(strlen(g_strstrip(sfile)) > 0)
		set_pref_file(g_strstrip(sfile));
	else
		set_pref_file(NULL);

	fclose (fp);

	return GRG_OK;
}
