/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_safe.c - security management for the program
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define __USE_GNU		//to use the global variable <environ> in stdlib.h

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <regex.h>
#include <unistd.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/fsuid.h>
#include <sys/types.h>

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <mcrypt.h>

#include "../libgringotts/libgringotts.h"

#include "gringotts.h"
#include "grg_defs.h"
#include "grg_utils.h"
#include "grg_safe.h"
#include "grg_widgets.h"

#ifdef PWD_HIDING
static guchar *session_key;
static gint session_key_len;
static guchar *session_IV;
static MCRYPT session_cipher;
#endif

static gboolean mem_safe = FALSE;

void
grg_security_init (void)
{
#ifdef PWD_HIDING
	session_cipher = mcrypt_module_open (SESSION_ALGO, NULL, MCRYPT_CFB, NULL);
	if(session_cipher == MCRYPT_FAILED)
		g_error("%s", _("Problem with libmcrypt, probably a faulty installation"));
	session_key_len = mcrypt_enc_get_key_size (session_cipher);
	session_key = grg_rnd_seq (gt, session_key_len);
	session_IV = grg_rnd_seq (gt, mcrypt_enc_get_iv_size (session_cipher));
#endif
}

#ifdef PWD_HIDING
gboolean
grg_security_seal (guchar *text, gint tlen)
{
	gint alg;
	alg = mcrypt_generic_init(session_cipher, session_key, session_key_len, session_IV);
	if (alg < 0){
		grg_msg(_("Problem with libmcrypt, probably a faulty installation"), GTK_MESSAGE_ERROR, NULL);
		return FALSE;
	}
	mcrypt_generic (session_cipher, text, tlen);
	mcrypt_generic_deinit (session_cipher);
	return TRUE;
}
#endif

#ifdef PWD_HIDING
gboolean
grg_security_unseal (guchar *text, gint tlen)
{
	gint alg;
	alg = mcrypt_generic_init(session_cipher, session_key, session_key_len, session_IV);
	if (alg < 0){
		grg_msg(_("Problem with libmcrypt, probably a faulty installation"), GTK_MESSAGE_ERROR, NULL);
		return FALSE;
	}
	mdecrypt_generic (session_cipher, text, tlen);
	mcrypt_generic_deinit (session_cipher);
	return TRUE;
}
#endif

void
grg_security_deinit (void)
{
#ifdef PWD_HIDING
	mcrypt_module_close (session_cipher);
	g_free(session_key);
	g_free(session_IV);
	session_key_len = 0;
#endif
}

gboolean
grg_security_filter (gboolean rootCheck)
{
	gint ulimit, canary;
	FILE *pipe;
#ifdef ENV_CHECK
	regex_t regexp1, regexp2;
	extern char **environ;
	gchar *lang, *display, *xauth;
#endif

	//drop eventual group root privileges
	setgid(getgid ());
	setgid(getgid ()); //twice for counter "saved IDs", cfr. Secure Programming HowTo
#ifdef HAVE_SYS_FSUID_H
	setfsgid(getgid ());
	setfsgid(getgid ());
#endif

	if (!geteuid())
	//the process is SUID root. I can mlockall() the memory in order to avoid swapping.
	{
		gint res = mlockall (MCL_FUTURE);

		if (res)
			g_critical ("%s", _("The process is setuid root, but I can't lock memory paging"));
		else
			mem_safe = TRUE;

		//drop root privileges
		setuid(getuid());
		setuid(getuid());
#ifdef HAVE_SYS_FSUID_H
		setfsuid(getuid());
		setfsuid(getuid());
#endif

#ifdef HAVE_SYS_FSUID_H
		if(getuid () && (!setuid(0) || !setfsuid(0)))
#else
		if(getuid () && !setuid(0))
#endif
		{
			g_critical ("%s", _("Cannot drop root privileges. Someone is hacking this process. I cannot go on"));
			return FALSE;
		}
	}

#ifdef HAVE_SYS_FSUID_H
	if(getuid () && (!setgid(0) || !setfsgid(0)))
#else
	if(getuid () && !setgid(0))
#endif
	{
		g_critical ("%s", _("Cannot drop root privileges. Someone is hacking this process. I cannot go on"));
		return FALSE;
	}

	if (!rootCheck && (!getuid () || !geteuid () || !getgid () || !getegid ()))
	//forbid root usage
	{
		g_critical ("%s", _("For security reasons, you cannot run Gringotts as root. Try using --enable-root-use in ./configure"));
		return FALSE;
	}

	//check core dump generation
	pipe = popen ("ulimit -Sc", "r");
	ulimit = fgetc (pipe);
	pclose (pipe);

	if (ulimit != '0')
	{
		g_critical ("%s", _("Your shell is configured to produce core dumps. I cannot go on. Please run gringotts via its shell script."));
		return FALSE;
	}

	//checks that stderr, stdin & stdout are opened
	canary = open("/dev/null", O_CREAT);
	if (canary < 3)
	{
		g_critical ("%s", _("stdin, stdout and/or stderr are invalid. Cannot continue."));
		close (canary);
		return FALSE;
	}

	//extract needed environmental vars, validate, erase environment,
	//and re-set them (see Secure Programming HowTo, sect.4.2)

#ifdef ENV_CHECK
	//extract
	lang = getenv ("LANG");
	display = getenv ("DISPLAY");
	xauth = getenv ("XAUTHORITY");
	//validate
	regcomp (&regexp1, "[[:alpha:]][[:alnum:]_,+@\\-\\.=]*",
		 REG_EXTENDED);
	regcomp (&regexp2,
		 "[[:digit:]{1,3}\\.[:digit:]{1,3}\\.[:digit:]{1,3}\\.[:digit:]{1,3}]?:[[:digit:]\\.]+",
		 REG_EXTENDED);
	if (((lang != NULL) && (regexec (&regexp1, lang, 0, NULL, 0)))
	    || ((xauth != NULL) && (!g_file_test (xauth, G_FILE_TEST_EXISTS)))
	    || (regexec (&regexp2, display, 0, NULL, 0)))
	{
		g_critical ("%s", _("Invalid environment variables. They could have been manipulated."));
		return FALSE;
	}
	//erase
	environ = NULL;
	//re-set (warning: don't free() the g_strconcat'ed strings)
	if (lang != NULL)
		putenv (g_strconcat ("LANG=", lang, NULL));
	putenv (g_strconcat ("DISPLAY=", display, NULL));
	if (xauth != NULL)
		putenv (g_strconcat ("XAUTHORITY=", xauth, NULL));
	putenv (g_strconcat ("HOME=", g_get_home_dir (), NULL));
#endif

	//this enables a stronger check on malloc() routines
	putenv ("MALLOC_CHECK_=2");

	return TRUE;
}

#define ADD_INDICATOR(box, text, pixbuf) \
	{ \
		GtkWidget *new_lbl, *img, *hbox; \
		new_lbl = gtk_label_new(text); \
		gtk_misc_set_alignment(GTK_MISC(new_lbl), 0, 0.5); \
		img = gtk_image_new_from_pixbuf (pixbuf); \
		hbox = gtk_hbox_new (FALSE, GRG_PAD); \
		gtk_box_pack_start(GTK_BOX(hbox), img, FALSE, FALSE, GRG_PAD); \
		gtk_box_pack_start(GTK_BOX(hbox), new_lbl, FALSE, FALSE, GRG_PAD); \
		gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, GRG_PAD); \
	}

const gchar *red_xpm[], *green_xpm[];

void
grg_security_monitor (void)
{
	GtkWidget *dialog = gtk_dialog_new_with_buttons (_("Security monitor"),
					      NULL,
					      GTK_DIALOG_MODAL |
					      GTK_DIALOG_DESTROY_WITH_PARENT,
						  GTK_STOCK_OK,
					      GTK_RESPONSE_OK,
					      NULL);
	GdkPixbuf *red = gdk_pixbuf_new_from_xpm_data (red_xpm);
	GdkPixbuf *green = gdk_pixbuf_new_from_xpm_data (green_xpm);
	
	if (geteuid () && getegid () && getuid () && getgid ())
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Running without root privileges"), green)
	else
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Running without root privileges"), red);
	
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from core dumps"), green);

	if(mem_safe)
	{
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from swap writings"), green);
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from ptrace spying"), green);
	}
	else
	{
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from swap writings"), red);
		ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Memory protection from ptrace spying"), red);
	}

#ifdef ENV_CHECK
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Environmental variables validation"), green);
#else
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Environmental variables validation"), red);
#endif

	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("stdout/stdin/stderr validation"), green);

#ifdef PWD_HIDING
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("In-memory password hiding"), green);
#else
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("In-memory password hiding"), red);
#endif

#ifdef STRONG_RANDOM
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Enforced use of /dev/random"), green);
#else
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Enforced use of /dev/random"), red);
#endif

#ifdef ROOT_FILTER
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Strict prohibition to root user"), green);
#else
	ADD_INDICATOR (GTK_DIALOG (dialog)->vbox, _("Strict prohibition to root user"), red);
#endif

	gtk_widget_show_all (dialog);
	gtk_dialog_run (GTK_DIALOG (dialog));
	g_object_unref (G_OBJECT(red));
	g_object_unref (G_OBJECT(green));
	gtk_widget_destroy (dialog);
}

/* XPM */
const gchar * red_xpm[] = {
"20 20 33 1",
"       c None",
".      c #1F1313",
"+      c #7F0A0A",
"@      c #ACA6A6",
"#      c #7A7A7A",
"$      c #CF0404",
"%      c #9D7676",
"&      c #461010",
"*      c #7E5E5E",
"=      c #DE7070",
"-      c #A00808",
";      c #675F5F",
">      c #AF5A5A",
",      c #630E0E",
"'      c #D65656",
")      c #E1A2A2",
"!      c #BB0606",
"~      c #F10202",
"{      c #472E2E",
"]      c #363232",
"^      c #8F0A0A",
"/      c #E10202",
"(      c #560E0E",
"_      c #AA2626",
":      c #D02222",
"<      c #6A2A2A",
"[      c #301212",
"}      c #710D0D",
"|      c #AE0606",
"1      c #862A2A",
"2      c #962A2A",
"3      c #BA2626",
"4      c #C3A6A6",
"      4>1+^2>)      ",
"    4<,}+^--||3)    ",
"   %&,}}+^-||!!$=   ",
"  #&(,}+^-||!$$$$=  ",
" @&((}++^-|!$$///$) ",
" {&&,,+^^|!!$//~~/: ",
"@[&&,}}^--!$$/~~~//)",
";[&((}+^^||$$/~~~/$'",
"][&&(,++--|!$//~//$:",
".[[((,++^-|!!$/$$$!!",
".[[&(,}+^--|!!$$$!!|",
"][[&((,++^-|||!!!!|_",
";.[&&(,}++^--|-|-|->",
"@..[&((,}++^^^----^)",
" ].[[&((,}++^+^^^^2 ",
" @.[[[&((,,}++++++4 ",
"  #..[&&((,,,}}}}%  ",
"   *.[[[&&((((,,%   ",
"    @][[[&&&((<4    ",
"      @;{[&{*@      "};

const gchar *green_xpm[] = {
"20 20 33 1",
"       c None",
".      c #131F13",
"+      c #0A7A0A",
"@      c #A6ACA6",
"#      c #07CC07",
"$      c #104310",
"%      c #A2E1A2",
"&      c #06A606",
"*      c #7A7A7A",
"=      c #105010",
"-      c #76A876",
";      c #769276",
">      c #70DE70",
",      c #0B8E0B",
"'      c #0F5F0F",
")      c #5F675F",
"!      c #02F102",
"~      c #5ABA5A",
"{      c #0E700E",
"]      c #56D656",
"^      c #08BD08",
"/      c #5A9A5A",
"(      c #02E102",
"_      c #5E7E5E",
":      c #122612",
"<      c #A6C3A6",
"[      c #0C9A0C",
"}      c #323632",
"|      c #05D705",
"1      c #0B850B",
"2      c #2E432E",
"3      c #123312",
"4      c #08B108",
"      </11,[~%      ",
"    <{'{1,,&&&^%    ",
"   ;='{+1[[44^^#>   ",
"  ;=''{1,[&4^####>  ",
" @3='{+1[&4^#||((|% ",
" 2$=''+1[&4##|!!((| ",
"@3$='{+,[&4#|(!!!(|%",
"):$=={+1[&^^|(!!!(|]",
"233='{+1[&4^#|(!(|##",
"::$=='{1,[&^#||||##^",
".:3$''{1,[&44^###^44",
"}:3$=''+1,[&44^4^444",
").33=='{+1,[&&&4&&&~",
"@.:3$=''{+1,,[[[[[[<",
" }:33$=''{+11,,,,,[ ",
" @.:3$==''{{+++1++< ",
"  *.:33===''{{{{{-  ",
"   ).:33$===''''-   ",
"    @}:33$$==={<    ",
"      @)23$2_@      "};

